"""
Contains build info, whether it's local built, or a pre-compiled wheel.
Auto-generated at compile time.
"""

build_info = {
    "version": "version",
    "build date": "build data",
    "python used": "python version",
    "platform": "platform"
}
